from collections.abc import Generator
from contextlib import contextmanager

from sqlalchemy import Engine
from sqlalchemy.orm import Session


@contextmanager
def session_factory(bind: Engine) -> Generator[Session, None, None]:
    with Session(bind=bind, expire_on_commit=False) as session:
        try:
            session.begin()
            yield session
        except:
            session.rollback()
            raise
        else:
            session.commit()



from typing import Any, ClassVar

from sqlalchemy.ext.asyncio import AsyncAttrs
from sqlalchemy.orm import DeclarativeBase as OriginalBase
from sqlalchemy.orm.mapper import Mapper
from typing_extensions import Self


class InzichtBase:
    __mapper__: ClassVar[Mapper]

    @classmethod
    def _get_primary_key(cls) -> list[str]:
        primary_key = [c.name for c in cls.__mapper__.primary_key]
        return primary_key

    @classmethod
    def _get_columns(cls) -> list[str]:
        columns = [c.name for c in cls.__mapper__.columns]
        return columns

    @classmethod
    def _get_relationships(cls) -> list[str]:
        relationships = [r.key for r in cls.__mapper__.relationships]
        return relationships

    @classmethod
    def _get_attributes(cls) -> list[str]:
        primary_key = set(cls._get_primary_key())
        columns = set(cls._get_columns())
        relationships = set(cls._get_relationships())
        attributes = columns | relationships
        safe_attributes = list(attributes - primary_key)
        return safe_attributes

    @classmethod
    def new(cls, **kwargs: Any) -> Self:
        safe_kwargs = {k: v for k, v in kwargs.items() if k in cls._get_attributes()}
        return cls(**safe_kwargs)

    def update(self, **kwargs: Any) -> None:
        safe_kwargs = {k: v for k, v in kwargs.items() if k in self._get_attributes()}
        for k, v in safe_kwargs.items():
            setattr(self, k, v)


class DeclarativeBase(AsyncAttrs, OriginalBase, InzichtBase):
    __abstract__ = True
    __mapper_args__ = {"eager_defaults": True}



from .aio.crud.factories import async_session_factory
from .aio.crud.generic import AioGenericCRUD
from .crud.factories import session_factory
from .crud.generic import GenericCRUD
from .declarative import DeclarativeBase, InzichtBase

try:
    from sqlmodel import SQLModel as OriginalSQLModel  # type: ignore[import-not-found]

    class SQLModel(OriginalSQLModel, InzichtBase):
        __mapper_args__ = {"eager_defaults": True}
except ImportError:
    pass


__all__ = [
    "DeclarativeBase",
    "GenericCRUD",
    "session_factory",
    "AioGenericCRUD",
    "async_session_factory",
    "SQLModel",
]



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession


@asynccontextmanager
async def async_session_factory(
    bind: AsyncEngine,
) -> AsyncGenerator[AsyncSession, None]:
    async with AsyncSession(bind=bind, expire_on_commit=False) as session:
        try:
            await session.begin()
            yield session
        except:
            await session.rollback()
            raise
        else:
            await session.commit()



from typing import Any


class BaseORMError(Exception):
    def __init__(self, message: str = "", **kwargs: Any):
        super().__init__(message)
        self.kwargs = kwargs


class DoesNotExistError(BaseORMError):
    pass


class IntegrityError(BaseORMError):
    pass


class UnknowError(BaseORMError):
    pass



from asyncio import Lock
from collections.abc import Generator, Sequence
from typing import Any, TypeVar, get_args

import sqlalchemy.exc
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession

from inzicht.aio.crud.interfaces import AioCRUDInterface
from inzicht.crud.errors import DoesNotExistError, IntegrityError, UnknowError
from inzicht.declarative import DeclarativeBase

T = TypeVar("T", bound=DeclarativeBase)


class AioGenericCRUD(AioCRUDInterface[T]):
    def __init__(self, async_session: AsyncSession) -> None:
        self.async_session = async_session
        self.lock = Lock()

    def get_model(self) -> type[T]:
        if hasattr(self, "__orig_class__"):
            (model,) = get_args(self.__orig_class__)
        elif hasattr(self, "__orig_bases__"):
            (base,) = self.__orig_bases__
            (model,) = get_args(base)
        else:
            raise TypeError(
                f"Can't define type parameter of generic class {self.__class__}"
            )
        return model

    async def count(self, where: Any | None = None) -> int:
        model = self.get_model()
        query = select(func.count()).select_from(model)
        if where is not None:
            query = query.filter(where)
        result = await self.async_session.execute(query)
        count = result.scalar() or 0
        return count

    async def create(self, instance: T | None = None, /, **kwargs: Any) -> T:
        model = self.get_model()
        if instance and kwargs:
            raise ValueError(
                "Cannot provide both 'instance' and keyword arguments for creation"
            )
        instance = instance or model.new(**kwargs)
        message = f"DB operation [CREATE] on instance '{instance}' of model '{model}'"
        try:
            async with self.lock:
                self.async_session.add(instance)
                await self.async_session.flush()
        except sqlalchemy.exc.IntegrityError as error:
            error_message = f"{message} failed because of integrity constraints"
            raise IntegrityError(error_message, **kwargs) from error
        except Exception as error:
            error_message = f"{message} failed because of unknown error"
            raise UnknowError(error_message, **kwargs) from error
        return instance

    async def bulk_create(self, instances: Sequence[T]) -> Sequence[T]:
        model = self.get_model()
        message = f"DB operation [BULK_CREATE] on {len(instances)} instances '[{instances[0]},...]' of model '{model}'"
        try:
            self.async_session.add_all(instances)
            await self.async_session.flush()
        except sqlalchemy.exc.IntegrityError as error:
            error_message = f"{message} failed because of integrity constraints"
            raise IntegrityError(error_message) from error
        except Exception as error:
            error_message = f"{message} failed because of unknow error"
            raise UnknowError(error_message) from error
        return instances

    async def get(self, id: int | str, /) -> T:
        model = self.get_model()
        instance = await self.async_session.get(model, id)
        message = f"DB operation [GET] on instance of model '{model}' with id '{id}'"
        if not instance:
            error_message = f"{message} failed because the instance was not found"
            raise DoesNotExistError(error_message, id=id)
        return instance

    async def read(
        self,
        *,
        where: Any | None = None,
        order_by: Any | None = None,
        skip: int = 0,
        take: int | None = None,
    ) -> Generator[T, None, None]:
        model = self.get_model()
        query = select(model)
        if where is not None:
            query = query.filter(where)
        if order_by is not None:
            query = query.order_by(order_by)
        if skip:
            query = query.offset(skip)
        if take:
            query = query.limit(take)
        result = await self.async_session.execute(query)
        items = (item for item in result.scalars())
        return items

    async def update(self, id: int | str, /, **kwargs: Any) -> T:
        model = self.get_model()
        instance = await self.async_session.get(
            model, id, with_for_update={"nowait": True}
        )
        message = f"DB operation [UPDATE] on instance of model '{model}' with id '{id}'"
        if not instance:
            error_message = f"{message} failed because the instance was not found"
            raise DoesNotExistError(error_message, id=id, **kwargs)
        instance.update(**kwargs)
        try:
            self.async_session.add(instance)
            await self.async_session.flush()
        except sqlalchemy.exc.IntegrityError as error:
            error_message = f"{message} failed because of integrity constraints"
            raise IntegrityError(error_message, **kwargs) from error
        except Exception as error:
            error_message = f"{message} failed because of unknow error"
            raise UnknowError(error_message, **kwargs) from error
        return instance

    async def delete(self, id: int | str, /) -> T:
        model = self.get_model()
        instance = await self.get(id)
        message = f"DB operation [DELETE] on instance {instance} of model '{model}' with id '{id}'"
        if not instance:
            error_message = f"{message} failed because the instance was not found"
            raise DoesNotExistError(error_message)
        await self.async_session.delete(instance)
        await self.async_session.flush()
        return instance
